// fdm81.c - finite-difference frequency domain cochlear model
//           Neely (1981, JASA 69, 1386-1393)

#include <stdio.h>
#include <math.h>

typedef struct {float x, y;} complex;

complex a[8][8][256], p[8][256], y[256], d[256];
float   w, p0p, srd, dx, dy, r1, r2, r4;

static float k0 = 1e9;
static float k1 = -2;
static float m0 = 0.15;
static float r0 = 200;
static float rho = 1;
static float tpi = 6.283183;
static float xl = 3.5;
static float yh = 0.1;
static int m = 8;
static int n = 246;

complex
cmplx(float x, float y)
{
    complex z;

    z.x = x;
    z.y = y;
    return (z);
}

complex
cneg(complex z)
{
    complex y;

    y.x = -z.x;
    y.y = -z.y;
    return (y);
}

complex
crecip(complex z)
{
    complex y;
    float   msq;

    msq = z.x * z.x + z.y * z.y;
    y.x = z.x / msq;
    y.y = -z.y / msq;
    return (y);
}

double
cabs(complex z)
{
    double   msq;

    msq = z.x * z.x + z.y * z.y;
    return (sqrt(msq));
}

complex
cadd(complex a, complex b)
{
    complex c;

    c.x = a.x + b.x;
    c.y = a.y + b.y;
    return (c);
}

complex
csub(complex a, complex b)
{
    complex c;

    c.x = a.x - b.x;
    c.y = a.y - b.y;
    return (c);
}

complex
cmul(complex a, complex b)
{
    complex c;

    c.x = a.x * b.x - a.y * b.y;
    c.y = a.x * b.y + a.y * b.x;
    return (c);
}

// mxint - initialize pressure and zero first sub-matrix

void
mxint(int m, int n)
{
    complex p1;
    int     i, j;

    p1 = cmplx(-2. * dx * p0p, 0.0);
    for (i = 0; i < m; i++) {
	p[i][0] = p1;
	p[i][n] = cmplx(0.0, 0.0);
	for (j = 0; j < m; j++)
	    a[i][j][0] = cmplx(0.0, 0.0);
    }
}

// mxadd - initialize the kth sub-matrix and add the previously inverted 
//	  (k-1)th sub-matrix:
//
//         a(k) = a(k) + a(k-1)

void
mxadd(int k, int m)
{
    int     i, m1, m2;

    m1 = m - 1;
    m2 = m - 2;
    a[0][0][k] = cadd(a[0][0][k], cmplx(r4 - y[k].y * srd, y[k].x * srd));
    a[0][1][k].x -= r2;
    for (i = 1; i < m1; i++) {
	a[i][i - 1][k].x -= r1;
	a[i][i][k].x += r4;
	a[i][i + 1][k].x -= r1;
    } 
    a[m1][m2][k].x -= r2;
    a[m1][m1][k].x += r4;
}

// mxinv - invert the kth sub-matrix of a(i,j) and solve for the pressure.

void
mxinv(int k, int m)
{
    complex at, b[8][8], q[8];
    int i, ii, j, ji, j1, mm;

    // work down from top of a(k)
    mm = m - 1;
    for (j = 0; j < m; j++) {
        if(cabs(a[j][j][k]) < 1e-9) {
	    fprintf(stderr, "*** Ill-conditioned matrix!\n");
            exit(1);
	}
        at = crecip(a[j][j][k]);		    // at = 1/a(i,j,k)
        for (j1 = 0; j1 < m; j1++) {
            a[j][j1][k] = cmul(a[j][j1][k], at);    // a(i,j1,k)=a(j,j1,k)*at
        }
        a[j][j][k] = at;
        if (j < mm) {
	    for (i = j + 1; i < m; i++) {
                at = a[i][j][k];
                a[i][j][k] = cmplx(0.0, 0.0);
	        for (j1 = 0; j1 < m; j1++) {
                    a[i][j1][k] = csub(a[i][j1][k], cmul(a[j][j1][k], at));
                }
            }
        }
    }
   // work up from bottom
    for (i = 0; i < m; i++) {
        for (j = 0; j < m; j++) {
            b[i][j] = a[i][j][k];
	}
     }
    for (i = 0; i < mm; i++) {
        for (j = i + 1; j < m; j++) {
            b[i][j] = cmplx(0.0, 0.0);
        }
    }
    for (ii = 0; ii < mm; ii++) {
        i = m - ii - 2;
	for (ji = i + 1; ji < m; ji++) {
	    for (j = 0; j < m; j++) {
                b[i][j] = csub(b[i][j], cmul(a[i][ji][k], b[ji][j]));
	    }
	}
    }

    // solve p(k) = b*p(k) and a(k) = -b
    for (i = 0; i < m; i++) {
        at = cmplx(0.0, 0.0);
	for (j = 0; j < m; j++) {
          at = cadd(at, cmul(p[j][k], b[i][j]));
          a[i][j][k] = cneg(b[i][j]);
        }
        q[i] = at;
    }
    for (i = 0; i < m; i++)
        p[i][k] = q[i];

}


// mxdbl - double the value of a(k)

void
mxdbl(int k, int m)
{
    int i, j;

    for (i = 0; i < m; i++)
        for (j = 0; j < m; j++)
            a[i][j][k] = cadd(a[i][j][k], a[i][j][k]);

}

// mxcpy - copy the (k-1)th submatrix to the kth submatrix

void
mxcpy(int k, int m)
{
    int i, j, km;

    km = k - 1;
    for (i = 0; i < m; i++) {
        p[i][k] = p[i][km];
        for (j = 0; j < m; j++) {
	    a[i][j][k] = a[i][j][km];
	}
    }

}

// mxelm - eliminate the upper-diagonal sub-matrices:
//
//         p(k) = p(k) - a(k)*p(k+1)

void
mxelm(int k, int m)
{
    complex ap;
    int i, j, kp;

    kp = k + 1;
    for (i = 0; i < m; i++) {
        ap = cmplx(0.0, 0.0);
	for (j = 0; j < m; j++) {
            ap = cadd(ap, cmul(a[i][j][k], p[j][kp]));
	}
        p[i][k] = csub(p[i][k], ap);
    }

}

// dbph - compute magnitude (in dB) and phase (unwrapped)

void
dbph(complex *p, int n)
{
    int i;
    double mg, db, ph, lp = -0.5;
    static double tpi = 2 * 3.1415927;
    
    for (i = 0; i < n; i++) {
        mg = hypot(p[i].x, p[i].y);
        if (mg > 1e-30) {
            db = 20 * log10(mg);
            ph = atan2(p[i].y, p[i].x) / tpi;
        } else {
            db = -600;
            ph = 0;
        }
        if (ph > lp) {
            ph -= floor(ph - lp + 0.5);
        } else {
            ph += floor(lp - ph + 0.5);
        }
        lp = ph;
        p[i].x = db;
        p[i].y = ph;
    }
}

//  wrsoln - write out data to disk

void
wrsoln(int ii, double f)
{
    char fn[80];
    complex tc;
    double dl;
    int i;
    FILE *fp;

    fprintf(stdout, "m=%d n=%d xl=%.2f yh=%.2f f=%.2f\n", m, n, xl, yh, f);
    sprintf(fn, "fdm81_%d.dat", ii);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - f=%.0f\n", fn, f);
    fprintf(fp, "; m=%d n=%d xl=%.2f yh=%.2f\n", m, n, xl, yh);

    /* BM displacement */
    tc = cmplx(0.0, -1 / w);
    for (i = 0; i < n; i++)
    	d[i] = cmul(cmul(y[i], p[0][i]),tc);

    /* convert to magnitude and phase */
    dbph(p[0], n);
    dbph(d, n);
    dbph(y, n);

    /* write data */
    fprintf(fp, "; x/L          P                 Dbm               Ybm     \n");
    fprintf(fp, ";         [dB]     [ph]     [dB]     [ph]     [dB]     [ph] \n");
    dl = 1.0 / (n - 1);
    for (i = 0; i < n; i++) {
        fprintf(fp, "%6.3f %8.1f %8.2f %8.1f %8.2f %8.1f %8.2f\n", i * dl,
        p[0][i].x, p[0][i].y, d[i].x, d[i].y, y[i].x, y[i].y);
    }
    fclose(fp);
}

int
main()
{
    float   ho, f, x;
    int     nm, k;

    dx = xl/(n-1);
    dy = yh/(m-1);
    r1 = (dx/dy)*(dx/dy);
    r2 = r1 + r1;
    r4 = 2. + r2;
    nm = n - 1;

    for (ho = 0; ho < 10; ho++) {	// loop over ten frequencies

	f = 400.0*pow(1.414, ho);
	w = tpi*f;
	srd = 4*w*rho*dy*r1;
	p0p = 2*w*w*rho;

	for (k = 0; k < n; k++) {	// compute admittance for all x
	    x = dx * k;
	    y[k] = crecip(cmplx(r0, m0 * w - k0 * exp(k1 * x) / w));
        }

	// solve block matrix equation

	mxint(m, n);
	mxadd(0, m);
	mxinv(0, m);
	mxdbl(0, m);
	for (k = 1; k < nm; k++) {		// work down from top
	    mxcpy(k, m);
	    mxadd(k, m);
	    mxinv(k, m);
	}

        // work up from bottom
        for (k = nm - 1;  k > 0; k--) {
	    mxelm(k, m);
        }
        wrsoln(ho, f);
    }

    return (0);
}
